#!/usr/bin/env python3

import subprocess
import argparse
import numpy as np
import sys
import matplotlib.pyplot as plt

parser = argparse.ArgumentParser(description=
#-------------------------------------------------------------------------
"""
Perform a block analysis of a column in a data file.
By default, the 1st column of data is analyzed (use --col to specify another column)
version 1.1      Written by Vincent Ballenegger (Feb. 2022)
""")

# Required arguments
parser.add_argument('file', type=str,
                    help='data file')

parser.add_argument('--col', type=int, default=1,
                    help='column number (counting from 1) to analyze')

#optional arguments
parser.add_argument('--start', type=int, default=1, 
                    help='nb of starting points to average over)')

parser.add_argument('--size', type=int, default=0, 
                    help='Block size. Enter 0 to scan over block sizes')

parser.add_argument('--minsize', type=int, default=1, 
                    help='shortest block length to consider (default: 1 observation/block)')

parser.add_argument('--maxsize', type=int, default=0, 
                    help='Largest block length to consider (default: N_observations/4)')

parser.add_argument("--plot", help="Shows plots (default: True)",
                    action="store_true", default=False)

parser.add_argument("--quick", help="Consider only a small number of block sizes (default: False)",
                    action="store_true", default=False)

parser.add_argument("--save", type=str, help="Save to file",
                    default="")

args = parser.parse_args()
#-------------------------------------------------------------------------

file_path = args.file
column = args.col
if column < 1:
    raise("Error: column number should be >= 1.")
nb_start_pts = args.start

X = np.loadtxt(file_path, usecols=(column-1))

Nobs         = len(X)           # total number of observations

minBlockSize = args.minsize
maxBlockSize = args.maxsize
if maxBlockSize == 0:
    maxBlockSize = int(Nobs/4)         # max: 4 blocs (otherwise can't calc variance)

if args.size != 0:
    maxBlockSize = minBlockSize = args.size
    args.plot = False
    #raise Exception("Option with a fixed block size not implemented yet !")

#---
def rcov(xj, i, Mi, Si):
        j  = i + 1
        Mj = Mi + (xj - Mi) / j
        Sj = Si + (i/j) * (xj - Mi) ** 2
        return (j, Mj, Sj)

def myvar(X):
        s = (0., 0, 0)
        for i in range(len(X)):
                s = rcov(X[i], s[0], s[1], s[2])
        return s[-1] / (len(X) - 1)                 # sample variance
#-------------------------------------------------------------------------------
def blockAverage(datastream):

    """
    Adapted from: https://gist.github.com/shane5ul/6289436ace018e6e9e79ee3489be2b8f

    by Vincent Ballenegger (March 2019)
    """

    NumBlocks = 1 + (maxBlockSize - minBlockSize)   # total number of block sizes

    blockMean = np.zeros((NumBlocks,nb_start_pts))               # mean (expected to be "nearly" constant)
    blockVar  = np.zeros((NumBlocks,nb_start_pts))               # variance associated with each blockSize
    blockCtr  = 0           #blockSize counter = index for referring to entries in arrays blockMean and blockVar
    
    block_mean = np.zeros(int(Nobs/minBlockSize))
    
    blockCtr = 0
    for blockSize in range(minBlockSize, maxBlockSize+1):   #+1 because the last value is not included in 'range()'

        #Consider only block sizes that divide the total number of observations
        # ignore also block sizes that are too large (would lead to less than 4 blocks)
        if args.quick and (Nobs % blockSize != 0 or Nobs/blockSize < 4):
            blockMean[blockCtr,0] = np.nan
            blockVar[blockCtr,0] = 0
            blockCtr += 1
            continue


        # if args.quick:
        #     if Nobs % blockSize > 1:   #skip block sizes with sub-optimal sizes
        #         blockMean[blockCtr,0] = np.nan
        #         blockVar[blockCtr,0] = 0
        #         blockCtr += 1
        #         continue

        Nblock    = int(Nobs/blockSize)               # total number of such blocks in datastream

        for start_pt in range(nb_start_pts):
            start_pos = start_pt * int(blockSize/nb_start_pts)

            block_mean.fill(0)
            # Loop to chop datastream into blocks and take average
            for i in range(1,Nblock+1):
                ibeg = (i-1) * blockSize + start_pos
                iend =  ibeg + blockSize

                #---------
                if iend <= Nobs:
                    block_data = datastream[ibeg:iend]
                    block_mean[i-1] = np.mean(block_data)
                else:
                    if ibeg >= Nobs:
                        ibeg -= Nobs
                        iend -= Nobs
                        block_data = datastream[ibeg:iend]
                        block_mean[i-1] = np.mean(block_data)
                    else:
                        block1 = datastream[ibeg:]
                        block2 = datastream[:blockSize-(Nobs-ibeg)]
                        # print('Nobs=', Nobs, 'blocksize=', blockSize)
                        # print('ibeg=', ibeg, 'iend=', iend)
                        # print(block1)
                        # print('size:', block1.size)
                        # print(block2)
                        # print('size:', block2.size)
                        # breakpoint()
                        if block1.size + block2.size != blockSize:
                            breakpoint()
                        block_mean[i-1] = ( np.mean(block1)*block1.size + np.mean(block2)*block2.size ) / blockSize

                # if iend < Nobs:
                #     block_data = datastream[ibeg:iend]
                # else:
                #     block_data = datastream.take(range(ibeg,iend), mode='wrap')
                #----
            blockMean[blockCtr,start_pt] = np.mean(block_mean[:Nblock])
            blockVar[blockCtr,start_pt]  = np.var(block_mean[:Nblock])/(Nblock - 1)
            #blockVar[blockCtr,start_pt]  = myvar(block_mean[:Nblock])/(Nblock - 1)
    
        #average over the starting points (we replace the first entry with the average !)
        blockMean[blockCtr,0] = np.mean(blockMean[blockCtr,:])
        blockVar[blockCtr,0] = np.mean(blockVar[blockCtr,:])
        
        blockCtr += 1

    block_sizes = np.arange(minBlockSize,maxBlockSize+1)

    #return the computed averages
    return block_sizes, blockVar[:,0], blockMean[:,0]

#--------------------------------------------------------------------------------
def str_with_err(value, error):
    """ formats a value with its uncertainty
    Examples:
    >>>print (str_with_err(17.969860, 5.966e-05))
    # 17.96986(6)
    >>> print (str_with_err(0.01202, 0.001749))
    # 0.012(2)
    """
    if error > 0:
        digits = -int(np.floor(np.log10(error)))
        #return "{0:.{2}f}({1:.0f})".format(value, error*10**digits, digits)
        return f"{value:.{2}f}({error*10**digits:.0f})"
    else:
        return f"{value:.{2}f}(UndefinedError)"
#--------------------------------------------------------------------------------

#Do the block analysis  

print(f"Analyzing {Nobs} values using {nb_start_pts} starting points")    

block_sizes, blockVar, blockMean = blockAverage(X)

if args.size:
    print("Number of blocks:", int(Nobs/args.size))

#Result using 1 starting point:
#print("<x> = {0:f} +/- {1:f}\n".format(blockMean[-1,0], np.sqrt(blockVar[-1,0])))

#Result for {nb_start_pts} starting point:
true_average = np.average(X)

max_var = np.amax(blockVar)             #choose the maximum variance for computing the uncertainty
#---- estimate the correlation time

#stddev = np.std(X)
#print "Estimated correlation time: ",  N*(sigma_bloc/stddev)**2 / 2
#----

uncertainty = np.sqrt(max_var)
print(f"<x>= {true_average} +/- {uncertainty} ,i.e. {str_with_err(true_average, uncertainty)}")
print(f"{true_average} {uncertainty}")

#--------------------------------------------------------------------------------

if args.plot:

    blockVar[blockVar == 0] = np.nan

    stride = (maxBlockSize - minBlockSize) /10

    plt.figure()
    #plt.subplot(2,1,1)

    #removes points with nan
    # input = np.hstack((block_sizes,np.sqrt(blockVar)))
    # print(input)
    # output = input[~np.isnan(input).any(axis=1)]
    # plt.plot(output[:,0], output[:,1],'r',marker='x',ls='-')
    block_sizes = block_sizes[~np.isnan(blockVar)]
    blockVar = blockVar[~np.isnan(blockVar)]

    plt.plot(block_sizes, np.sqrt(blockVar),'r',marker='x',ls='-')

    # if args.quick:
    #     plt.plot(block_sizes, np.sqrt(blockVar),'r-',marker='x')
    # else:
    #     plt.plot(block_sizes, np.sqrt(blockVar),'r-',lw=1)

    plt.xlabel('block size')
    plt.xscale('log')
    plt.ylabel(r'$\sigma_{{\rm block\ av.} / \sqrt{n_{\rm blocks}}} \;=\; \sigma_{\overline{x}}$')
    plt.plot(0,0)
    plt.margins(x=0)
    plt.margins(y=0)

    #plt.subplot(2,1,2)
    #plt.plot([0,maxBlockSize], [true_average, true_average],c='k')
    ##plt.errorbar(v[::stride], blockMean[::stride,0], np.sqrt(blockVar[:,0])[::stride])
    #plt.plot(block_sizes,  blockMean, c='r',marker='x')
    #plt.ylabel('<x>')
    #plt.xlabel('block size')

    plt.tight_layout()

    if len(args.save) != 0:
        plt.savefig(args.save)
    plt.show()